if select(2, UnitClass("player")) ~= "PRIEST" then return end

local name, addon = ...
_G[name] = addon

local LSM = LibStub("LibSharedMedia-3.0")
local LAD = LibStub("LibArtifactData-1.0")
local pairs, ipairs, select = pairs, ipairs, select
local format, gsub, sqrt = format, gsub, sqrt
local UnitBuff, UnitPower, GetSpellCooldown = UnitBuff, UnitPower, GetSpellCooldown
local GetNetStats, GetTime = GetNetStats, GetTime

local vilins = "|cFF9370DB" .. select(2, GetAddOnInfo(name)) .. "|r"
addon.Vilins = vilins

-- Shadow priest
local priestClass, shadowSpec, powerToken, powerType, insanity = "PRIEST", SPEC_PRIEST_SHADOW, "INSANITY", 13, 0
-- Player
local playerSpellHaste = 0
-- Spell ids
local mbId, swvId, vtId, mfId = 8092, 205351, 34914, 15407
-- Buffs
local vfName, liName = GetSpellInfo(194249), GetSpellInfo(197937)
local isVoiTActive
-- Insanity multipliers
local piMult, stmMult, fotmMult = 1, 1, 1
-- Voidform
local vfCurrStackTime, vfStacks, vfDrainStacks, vfThreshold, vfDrainMod = 0, 0, 0, 100, 1
-- Talents
local isLotvSelected, lotvThreshold, isAsSelected
-- Text colors
local insHexColor, castHexColor, asHexColor
-- AS
local asTargets = {}
-- Call to the Void
local cttv, cttvActive = {}, false
-- LI
local liStacks, displayLiStacks = 0
-- Insanity gain
local castGain, asGain, cttvGain = 0, 0, 0
-- VF reports
local vfRep, vfDisplayRepEnabled, vfSaveRepEnabled, maxSavedReps, currEncName = {}
-- GCD --
local gcdStart, gcdDuration, gcdBarEnable, gcdBarDeplete = 0, 0
-- Frames
local frame = CreateFrame("Frame", name .. "MainFrame", UIParent)
frame:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)
local bgFrame = CreateFrame("Frame", name .. "Background", frame)
local bordFrame = CreateFrame("Frame", name .. "Border", bgFrame)
local bordVfFrame = CreateFrame("Frame", name .. "BorderVf", bordFrame)
local insBar = CreateFrame("StatusBar", name .. "InsBar", bgFrame)
local castBar, asBar
local vfMarkFrame = CreateFrame("Frame", name .. "VfMark", insBar)
local gcdBar = CreateFrame("StatusBar", name .. "GcdBar", bgFrame)
-- Font strings
local insFontString = insBar:CreateFontString(nil, "OVERLAY")
local vfStackFontString = insBar:CreateFontString(nil, "OVERLAY")
local vfTimeFontString = insBar:CreateFontString(nil, "OVERLAY")
-- Animations
local oocFadeAnim = frame:CreateAnimationGroup()
local vfReadyAnim = bordVfFrame:CreateAnimationGroup()
local vfEndAnim = bordVfFrame:CreateAnimationGroup()
-- DB
local db, defaults
-- T20 Set --
local t20SetItemIds = { 147163, 147164, 147165, 147166, 147167, 147168 }
local isT20Set2pActive, isT20Set4pActive = false, false
local emName, emStacks = GetSpellInfo(247226), 0
local t20SetDrainMod, t20SetStmDrainMod = 0.90, 0.95

addon.frame, addon.oocFadeAnim = frame, oocFadeAnim

local function InitFrames()
	bgFrame:SetAllPoints()
	bgFrame.tex = bgFrame:CreateTexture(nil, "BACKGROUND")
	bgFrame.tex:SetAllPoints()
	bgFrame.tex:SetColorTexture(1, 1, 1, 1)
	
	bordFrame:SetAllPoints()
	bordFrame.texs = {}
	bordVfFrame:SetAllPoints(bordFrame)
	bordVfFrame.texs = {}
	for i=1,4,1 do
		bordFrame.texs[i] = bordFrame:CreateTexture()
		bordFrame.texs[i]:SetColorTexture(1, 1, 1, 1)
		bordVfFrame.texs[i] = bordVfFrame:CreateTexture()
		bordVfFrame.texs[i]:SetColorTexture(1, 1, 1, 1)
	end
	
	insBar:SetAllPoints()
	castBar = insBar:CreateTexture()
	asBar = insBar:CreateTexture()
	
	gcdBar.tex = gcdBar:CreateTexture(nil, "BACKGROUND")
	gcdBar.tex:SetAllPoints()
	gcdBar.tex:SetColorTexture(1, 1, 1, 1)
	
	vfMarkFrame.tex = vfMarkFrame:CreateTexture()
	vfMarkFrame.tex:SetAllPoints()
	vfMarkFrame.tex:SetColorTexture(1, 1, 1, 1)
	
	insFontString:SetPoint("RIGHT", insBar, "RIGHT", -2, 0)
	vfTimeFontString:SetPoint("LEFT", insBar, "LEFT", 2, 0)
	vfStackFontString:SetPoint("CENTER", insBar, "CENTER")
	
	oocFadeAnim:SetToFinalAlpha(true)
	oocFadeAnim.alpha = oocFadeAnim:CreateAnimation("Alpha")
	oocFadeAnim.alpha:SetFromAlpha(1.0)
	oocFadeAnim.alpha:SetSmoothing("NONE")
	
	vfReadyAnim:SetToFinalAlpha(true)
	vfReadyAnim:SetLooping("REPEAT")
	vfReadyAnim.alphaIn = vfReadyAnim:CreateAnimation("Alpha")
	vfReadyAnim.alphaIn:SetSmoothing("OUT")
	vfReadyAnim.alphaIn:SetOrder(1)
	vfReadyAnim.alphaIn:SetFromAlpha(0.0)
	vfReadyAnim.alphaIn:SetToAlpha(1.0)
	vfReadyAnim.alphaOut = vfReadyAnim:CreateAnimation("Alpha")
	vfReadyAnim.alphaOut:SetSmoothing("IN")
	vfReadyAnim.alphaOut:SetOrder(2)
	vfReadyAnim.alphaOut:SetFromAlpha(1.0)
	vfReadyAnim.alphaOut:SetToAlpha(0.0)
	
	vfEndAnim:SetToFinalAlpha(true)
	vfEndAnim.alpha = vfEndAnim:CreateAnimation("Alpha")
	vfEndAnim.alpha:SetSmoothing("NONE")
	vfEndAnim.alpha:SetFromAlpha(1.0)
	vfEndAnim.alpha:SetToAlpha(0.0)
	vfEndAnim.alpha:SetDuration(0.1)
end

local function SetInsanityGainBarValue(bar, baseValue, value)
	local parent = bar:GetParent()
	local _, maxInsanity = parent:GetMinMaxValues()
	local maxValue = baseValue + value
	if maxValue > maxInsanity then
		value = maxInsanity - baseValue
	end
	local width = parent:GetWidth()
	local left = baseValue / maxInsanity
	local right = (baseValue + value) / maxInsanity
	local leftPosition = (baseValue / maxInsanity) * width
	local barWidth = (value / maxInsanity) * width
	if barWidth < 0.5 then
		bar:Hide()
	else
		bar:ClearAllPoints()
		bar:SetTexCoord(left, right, 0, 1)
		bar:SetWidth(barWidth)
		bar:SetPoint("TOPLEFT", leftPosition, 0)
		bar:SetPoint("BOTTOMLEFT", leftPosition, 0)
		bar:Show()
	end
end

function frame:ClearSavedVfReps()
	for i = 1,#db.vfreportarchive do
		table.remove(db.vfreportarchive)
	end
	print(vilins .. " VF report:")
	print("Saved reports cleared!")
end

local function GetInsanityDrain(stack)
	return 6 + ((stack - 1) * (2/3) * vfDrainMod)
end

local function PrintVfRep(report, fromArchive)
	if (vfDisplayRepEnabled == 1 and report.stmduration ~= nil) or vfDisplayRepEnabled == 2 or fromArchive == true then
		print(format(vilins .. " VF report: %s", date("%X %x", report.time)))
		if report.encounter ~= nil then
			print("Encounter: " .. report.encounter)
		end
		
		print(format("VF duration: %.1fs", report.duration))
		if report.stmduration ~= nil then
			print(format("STM duration: %.1fs", report.stmduration))
		end
		print(format("VF stacks: %d", report.vfstacks))
		print(format("VF drain stacks (drain): %d (%.1f)", report.vfdrainstacks,  GetInsanityDrain(report.vfdrainstacks)))
	end
end

local function PrintSavedVfReps()
	if #db.vfreportarchive == 0 then
		print(vilins .. " VF report:")
		print("You currently have no saved reports.")
		return
	end
	print("------------------------")
	for _,v in pairs(db.vfreportarchive) do
		PrintVfRep(v, true)
		print("------------------------")
	end
end

local function SaveVfRep(report)
	if (vfSaveRepEnabled == 1 and report.stmduration ~= nil) or vfSaveRepEnabled == 2 then
		local size = #db.vfreportarchive
		local copyRep = {}
		for orig_key, orig_value in pairs(report) do
            copyRep[orig_key] = orig_value
        end
		table.insert(db.vfreportarchive, copyRep)
		local d = size - maxSavedReps
		if d >= 0 then
			for i = 1,(d+1) do
				table.remove(db.vfreportarchive, 1)
			end
		end
	end
end

local function CreateVfRep()
	vfRep.time = time()
	vfRep.duration = GetTime()
	vfRep.encounter = currEncName
	vfRep.vfstacks = 0
	vfRep.vfdrainstacks = 0
	vfRep.vfdrain = 0
end

local function FinaliseVfRep()
	local currentTime = GetTime()
	vfRep.duration = currentTime - vfRep.duration
	if vfRep.stmduration then
		vfRep.stmduration = currentTime - vfRep.stmduration
	end
	vfRep.vfdrain = GetInsanityDrain(vfRep.vfdrainstacks)
	
	PrintVfRep(vfRep, false)
	SaveVfRep(vfRep)
	
	vfRep.stmduration = nil
end

local function GetVoidformTime()
	if insanity > 0 and vfStacks > 0 then
		-- Drain current stack
		local vfCurrStackTimeLeft = 1.0 - (GetTime() - vfCurrStackTime)-- + select(4, GetNetStats()) / 1000
		local vfCurrStackDrain = GetInsanityDrain(vfDrainStacks)
		if ((vfCurrStackDrain * vfCurrStackTimeLeft) > insanity) then
			vfCurrStackTimeLeft = insanity / vfCurrStackDrain
			return vfCurrStackTimeLeft
		else
			local ins = insanity - (vfCurrStackDrain * vfCurrStackTimeLeft)
			-- Sum d+(x-1)*f*dm, Sum x=m to n => (-1/2)*(m-n-1)*(2*d+f*dm*(m+n-2))=ins => Solve for n, d=6, f=(2/3)
			local dm, m = vfDrainMod, vfDrainStacks + 1 -- We already drained current stack
			local vfRemStacks = (sqrt(dm*dm*(9-12*m+4*m*m)+dm*(72*m-108+12*ins)+324)+dm-18) / (2*dm)
			return vfCurrStackTimeLeft + (vfRemStacks - vfDrainStacks)
		end
	else
		return 0
	end
end

local function UpdateGcd(spell)
	local start, duration = GetSpellCooldown(spell)
	if start and start > 0 and gcdBarEnable then
		gcdStart = start
		gcdDuration = duration
		gcdBar:Show()
	end
end

local function OnUpdate(self)
	local currTime = GetTime()
	
	insBar:SetValue(insanity)
	
	cttvGain = 0
	for guid in pairs(cttv) do
		if cttv[guid] ~= nil then
			if (currTime - cttv[guid].spawnTime) > 10 then
				cttv[guid] = nil
			else
				if cttv[guid].isCasting then
					cttvGain = cttvGain + 3 * fotmMult * piMult * stmMult
				end
			end
		end
	end
	local totalCastGain = castGain + cttvGain
	SetInsanityGainBarValue(castBar, insanity, totalCastGain)
	
	asGain = 0
	for guid in pairs(asTargets) do
		if asTargets[guid] ~= nil then
			if (currTime - asTargets[guid].lastUpdateTime) > 10 then
				asTargets[guid] = nil
			else
				asGain = asGain + asTargets[guid].asCount*3 * piMult * stmMult
			end
		end
	end
	SetInsanityGainBarValue(asBar, totalCastGain + insanity, asGain)
	
	if isT20Set4pActive then
		if stmMult > 1 then
			vfDrainMod = t20SetStmDrainMod
		else
			vfDrainMod = t20SetDrainMod
		end
	else
		vfDrainMod = 1
	end
	
	local castGainText = (castGain + cttvGain) > 0 and "|c" .. castHexColor .. castGain + cttvGain .. "|r + " or ""
	local asGainText = asGain > 0 and "|c" .. asHexColor .. asGain .. "|r + " or ""
	insFontString:SetText(asGainText .. castGainText .. insanity)
	if (vfStacks > 0) then
		vfStackFontString:SetText(format("%d (%d%%)", vfStacks, playerSpellHaste))
		vfTimeFontString:SetText(format("%.1f (%d)", GetVoidformTime(), vfDrainStacks))
	elseif (liStacks > 0) and displayLiStacks then
		vfStackFontString:SetText(format("%d", liStacks))
	end
	
	if gcdStart > 0 then
		local pct = (currTime - gcdStart) / gcdDuration
		if pct > 1 then
			gcdBar:SetValue(0)
			gcdBar:Hide()
			gcdStart = 0
		else
			if gcdBarDeplete then
				gcdBar:SetValue(1-pct)
			else
				gcdBar:SetValue(pct)
			end
		end
	end
end

-- Return 2pc, 4pc
local function IsSetBonusActive(setItemIds)
	local setItems = 0
	for i = 1, #setItemIds do
		if IsEquippedItem(setItemIds[i]) then
			setItems = setItems + 1
		end
	end
	if setItems >= 4 then
		return true, true
	elseif setItems >= 2 then
		return true, false
	else
		return false, false
	end
end

frame:RegisterEvent("ADDON_LOADED")
function frame:ADDON_LOADED(...)
	if name == ...	then
		self.db = LibStub("AceDB-3.0"):New("VilinkasInsanityDB", default, true)
		db = self.db.profile
		InitFrames()
		self:UpdateOptions()
		
		self:ACTIVE_TALENT_GROUP_CHANGED()
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	end
end

function frame:ACTIVE_TALENT_GROUP_CHANGED()
	if GetSpecialization() == SPEC_PRIEST_SHADOW then
		self:RegisterEvent("PLAYER_TALENT_UPDATE")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("ENCOUNTER_START")
		self:RegisterEvent("ENCOUNTER_END")
		self:RegisterUnitEvent("UNIT_SPELLCAST_START", "player")
		self:RegisterUnitEvent("UNIT_SPELLCAST_SUCCEEDED", "player")
		self:RegisterUnitEvent("UNIT_SPELLCAST_STOP", "player")
		self:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_START", "player")
		self:RegisterUnitEvent("UNIT_SPELLCAST_CHANNEL_STOP", "player")
		self:RegisterUnitEvent("UNIT_POWER_FREQUENT", "player")
		self:RegisterUnitEvent("UNIT_MAXPOWER", "player")
		self:RegisterUnitEvent("UNIT_SPELL_HASTE", "player")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:RegisterUnitEvent("UNIT_INVENTORY_CHANGED", "player")
		
		self:SetScript("OnUpdate", OnUpdate)
		
		if not UnitAffectingCombat("player") then
			local oocAlpha = oocFadeAnim.alpha:GetToAlpha()
			self:SetAlpha(oocAlpha)
		end
		
		self:UNIT_INVENTORY_CHANGED()
		
		self:Show()
	else
		self:UnregisterEvent("PLAYER_TALENT_UPDATE")
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("ENCOUNTER_START")
		self:UnregisterEvent("ENCOUNTER_END")
		self:UnregisterEvent("UNIT_SPELLCAST_START")
		self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:UnregisterEvent("UNIT_SPELLCAST_STOP")
		self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START")
		self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
		self:UnregisterEvent("UNIT_POWER_FREQUENT")
		self:UnregisterEvent("UNIT_MAXPOWER")
		self:UnregisterEvent("UNIT_SPELL_HASTE")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
		
		self:SetScript("OnUpdate", nil)
		
		self:Hide()
	end
end

function frame:PLAYER_TALENT_UPDATE()
	fotmMult = 1
	if select(4, GetTalentInfo(1, 2, 1)) then
		fotmMult = 1.2
	end
	isAsSelected = select(4, GetTalentInfo(5, 2, 1))
	if isAsSelected == false then
		asTargets = {}
	end
	if select(4, GetTalentInfo(7, 1, 1)) then
		vfMarkFrame:Show()
		vfThreshold = lotvThreshold
	else
		vfMarkFrame:Hide()
		vfThreshold = 100
	end
		
	self:UNIT_MAXPOWER()
end

function frame:ARTIFACT_EQUIPPED_CHANGED(message, newArtifactID, oldArtifactID)
	cttvActive = false
	if newArtifactID == 128827 then
		local id, data = LAD:GetArtifactTraits(newArtifactID)
		for _,v in pairs(data) do
			if v.traitID == 1575 then
				cttvActive = true
			end
		end
	end
end
LAD.RegisterCallback(frame, "ARTIFACT_EQUIPPED_CHANGED")

function frame:ARTIFACT_TRAITS_CHANGED(message, artifactID, traitsData)
	cttvActive = false
	if artifactID == 128827 then
		for _,v in pairs(traitsData) do
			if v.traitID == 1575 then
				cttvActive = true
			end
		end
	end
end
LAD.RegisterCallback(frame, "ARTIFACT_TRAITS_CHANGED")

function frame:UNIT_INVENTORY_CHANGED()
	isT20Set2pActive, isT20Set4pActive = IsSetBonusActive(t20SetItemIds)
end

function frame:PLAYER_ENTERING_WORLD()
	if not UnitAffectingCombat("player") then
		local oocAlpha = oocFadeAnim.alpha:GetToAlpha()
		self:SetAlpha(oocAlpha)
	end
	asTargets = {}
	self:UNIT_POWER_FREQUENT()
end

function frame:PLAYER_REGEN_ENABLED()
	oocFadeAnim:Play()
end

function frame:PLAYER_REGEN_DISABLED()
	oocFadeAnim:Stop()
	self:SetAlpha(1)
end

local canEnterVfPrev = false
function frame:UNIT_POWER_FREQUENT()
	insanity = UnitPower("player", powerType)
	if not (vfStacks > 0) then
		if (insanity >= vfThreshold) and not canEnterVfPrev then
			vfEndAnim:Stop()
			vfReadyAnim:Play()
			canEnterVfPrev = true
		elseif insanity < vfThreshold and canEnterVfPrev then
			vfReadyAnim:Finish()
			canEnterVfPrev = false
		end
	end
end

function frame:UNIT_MAXPOWER()
	insBar:SetMinMaxValues(0, UnitPowerMax("player", powerType))
end

function frame:UNIT_SPELL_HASTE(unitId)
	playerSpellHaste = UnitSpellHaste(unitId)
end

function frame:UNIT_SPELLCAST_START(unitID, spell, rank, lineID, spellID)
	if spellID == mbId then
		castGain = (15 + emStacks) * fotmMult
	elseif spellID == swvId then
		castGain = 25
	elseif spellID == vtId then
		castGain = 6
	end
	castGain = castGain * piMult * stmMult
	
	if not UnitAffectingCombat("player") and castGain > 0 then
		oocFadeAnim:Stop()
		self:SetAlpha(1)
	end
	
	UpdateGcd(spell)
end

function frame:UNIT_SPELLCAST_SUCCEEDED(unitID, spell)
	UpdateGcd(spell)
end

function frame:UNIT_SPELLCAST_STOP(unitID, spell)
	if not UnitAffectingCombat("player") and castGain > 0 then
		oocFadeAnim:Play()
	end
	castGain = 0
end

function frame:UNIT_SPELLCAST_CHANNEL_START(unitID, spell, rank, lineID, spellID)
	if spellID == mfId then
		castGain = 3 * fotmMult
	end
	castGain = castGain * piMult * stmMult
end

function frame:UNIT_SPELLCAST_CHANNEL_STOP(unitID, spell)
	castGain = 0
end

function frame:ENCOUNTER_START(encounterID, encounterName)
	currEncName = encounterName
end

function frame:ENCOUNTER_END(encounterID, encounterName)
	currEncName = nil
end

function frame:COMBAT_LOG_EVENT_UNFILTERED(_, event, _, sourceGUID, _, _, _, destGUID, _, _, _, spellId, spellName)
	--print("COMBAT_LOG_EVENT: event: %s, spell id: %s, spell name: %s", event, spellId, spellName)
	if sourceGUID == UnitGUID("player") then
		-- Voidform --
		if spellId == 194249 then
			if event == "SPELL_AURA_APPLIED" then
				vfCurrStackTime = GetTime()
				vfDrainStacks = 1
				vfStacks = select(4, UnitBuff("player", vfName))
				vfEndAnim:Stop()
				vfReadyAnim:Stop()
				bordVfFrame:SetAlpha(1.0)
				CreateVfRep()
			elseif event == "SPELL_AURA_APPLIED_DOSE" then
				vfCurrStackTime = GetTime()
				if not isVoiTActive then
					vfDrainStacks = vfDrainStacks + 1
					vfRep.vfdrainstacks = vfDrainStacks
				end
				vfStacks = vfStacks + 1
				vfRep.vfstacks = vfStacks
			elseif event == "SPELL_AURA_REMOVED" then
				vfDrainStacks = 0
				vfStacks = 0
				vfCurrStackTime = 0
				vfEndAnim:Play()
				vfStackFontString:SetText("")
				vfTimeFontString:SetText("")
				FinaliseVfRep()
			end
		-- Lingering Insianity --
		elseif spellId == 197937 then
			if event == "SPELL_AURA_APPLIED" then
				liStacks = select(4, UnitBuff("player", liName))
			elseif event == "SPELL_AURA_REMOVED_DOSE" then
				liStacks = liStacks - 2
			elseif event == "SPELL_AURA_REMOVED" then
				liStacks = 0
				if vfStacks < 1 then
					vfStackFontString:SetText("")
				end
			end
		-- Power Infusion --
		elseif spellId == 10060 then
			if event == "SPELL_AURA_APPLIED" then
				piMult = 1.25
			elseif event == "SPELL_AURA_REMOVED" then
				piMult = 1
			end
		-- Surrender to Madness --
		elseif spellId == 193223 then
			if event == "SPELL_AURA_APPLIED" then
				stmMult = 2
				vfRep.stmduration = GetTime()
			elseif event == "SPELL_AURA_REMOVED" then
				stmMult = 1
				FinaliseVfRep()
			end
		-- Auspicious Spirits Spawn --
		elseif spellId == 147193 then
			if event == "SPELL_CAST_SUCCESS" then
				if isAsSelected then
					if not asTargets[destGUID] then
						asTargets[destGUID] = {}
						asTargets[destGUID].asCount = 0
					end
					asTargets[destGUID].asCount = asTargets[destGUID].asCount + 1
					asTargets[destGUID].lastUpdateTime = GetTime()
				end
			end
		-- Auspicious Spirits Despawn --
		elseif spellId == 148859 then
			if event == "SPELL_DAMAGE" then
				if isAsSelected then
					if asTargets[destGUID] then
						asTargets[destGUID].asCount = asTargets[destGUID].asCount - 1
						asTargets[destGUID].lastUpdateTime = GetTime()
						if asTargets[destGUID].asCount <= 0 then
							asTargets[destGUID] = nil
						end
					end
				end
			end
		-- Void Torrent --
		elseif spellId == 205065 then
			if event == "SPELL_AURA_APPLIED" then
				isVoiTActive = true
			elseif event == "SPELL_AURA_REMOVED" then
				isVoiTActive = false
			end
		-- Call to the Void --
		elseif spellId == 193470 then
			if event == "SPELL_SUMMON" then
				if cttvActive then
					cttv[destGUID] = {}
					cttv[destGUID].spawnTime = GetTime()
				end
			end
		-- Empty Mind --
		elseif spellId == 247226 then
			if event == "SPELL_AURA_APPLIED" then
				emStacks = select(4, UnitBuff("player", emName))
			elseif event == "SPELL_AURA_APPLIED_DOSE" then
				emStacks = emStacks + 1
			elseif event == "SPELL_AURA_REMOVED" then
				emStacks = 0
			end
		end
	-- Void Tentacle spawned by Call to the Void --
	elseif cttv[sourceGUID] then
		-- Void Tentacle's Mind Flay --
		if spellId == 193473 then
			if event == "SPELL_AURA_APPLIED" then
				cttv[sourceGUID].isCasting = true
			elseif event == "SPELL_AURA_REMOVED" then
				cttv[sourceGUID].isCasting = false
			end
		end
	end
	if event == "UNIT_DIED" then
		if destGUID == UnitGUID("player") then
			if stmMult > 1 or vfStacks > 0 then
				FinaliseVfRep()
			end
			asTargets = {}
			vfCurrStackTime = 0
			vfDrainStacks = 0
			vfStacks = 0
			liStacks = 0
			piMult = 1
			stmMult = 1
			isVoiTActive = false
			vfReadyAnim:Stop()
			vfEndAnim:Play()
			insFontString:SetText("")
			vfStackFontString:SetText("")
			vfTimeFontString:SetText("")
		else
			if asTargets[destGUID] then
				asTargets[destGUID] = nil
			end
		end
	end
end

default = {
	profile = {
		general = {
			posx = 0,
			posy = 0,
			width = 250,
			height = 22,
			alphaooc = 0.3,
			orientation = "HORIZONTAL",
			anchor = "SCREEN",
		},
		bars = {
			texture = "Blizzard Raid Bar",
			insbarcolor = { 0.4, 0, 0.8, 1 },
			castbarcolor = { 1, 1, 1, 1 },
			asbarcolor = { 1, 0.31, 0.85, 1 },
		},
		background = {
			color = { 0.2, 0.2, 0.2, 0.5 },
		},
		border = {
			thickness = 1,
			color = { 0.0, 0.0, 0.0, 0.7 },
		},
		overlay = {
			thickness = 1,
			color = { 0, 0, 0, 1 },
		},
		text = {
			file = "Friz Quadrata TT",
			flag = "OUTLINE",
			insanitycolor = { 1, 1, 1, 1 },
			insanitysize = 12,
			castcolor = { 1, 1, 1, 1 },
			ascolor = { 1, 0.31, 0.85, 1 },
			vfstackcolor = { 1, 1, 1, 1 },
			vfstacksize = 12,
			vftimecolor = { 1, 1, 1, 1 },
			vftimesize = 12,
		},
		misc = {
			lotvthreshold = 65,
			lingeringinsanity = true,
			vfdisreportingenabled = 2,
			vfsavereportingenabled = 1,
			vfreportingmaxsaved = 3,
			--patch70105enable = false,
		},
		animations = {
			oocduration = 0.5,
			oocstartdelay = 1,
			bordercolor = { 0.7, 0.0, 1.0, 1.0 },
			borderduration = 1,
		},
		gcd = {
			enable = true,
			pos = "TOP",
			height = 5.0,
			barcolor = { 1, 1, 1, 1 },
			deplete = false,
		},
		vfreportarchive = {
		},
	},
}

local function SetBorderSize(border, size)
	for i=1,4,1 do
		border.texs[i]:ClearAllPoints()
	end
	border.texs[1]:SetPoint("TOPRIGHT", border, "TOPLEFT", 0, size)
	border.texs[1]:SetPoint("BOTTOMRIGHT", border, "BOTTOMLEFT", 0, -size)
	border.texs[1]:SetSize(size, 0)
	border.texs[2]:SetPoint("TOPLEFT", border, "TOPRIGHT", 0, size)
	border.texs[2]:SetPoint("BOTTOMLEFT", border, "BOTTOMRIGHT", 0, -size)
	border.texs[2]:SetSize(size, 0)
	border.texs[3]:SetPoint("TOPRIGHT", border, "BOTTOMLEFT")
	border.texs[3]:SetPoint("TOPLEFT", border, "BOTTOMRIGHT")
	border.texs[3]:SetSize(0, size)
	border.texs[4]:SetPoint("BOTTOMRIGHT", border, "TOPRIGHT")
	border.texs[4]:SetPoint("BOTTOMLEFT", border, "TOPLEFT")
	border.texs[4]:SetSize(0, size)
end

-- Conversion function from RGBA to HEX, Credit: https://gist.github.com/marceloCodget/3862929
local function RGBAToHex(rgba)
	local hexadecimal = ''
	-- Convert to wow supported format AARRGGBB
	local argb = { rgba[4]*255, rgba[1]*255, rgba[2]*255, rgba[3]*255 }
	for key, value in pairs(argb) do
		local hex = ''
		while(value > 0)do
			local index = math.fmod(value, 16) + 1
			value = math.floor(value / 16)
			hex = string.sub('0123456789ABCDEF', index, index) .. hex			
		end
		if(string.len(hex) == 0) then 
			hex = '00'
		elseif(string.len(hex) == 1)then
			hex = '0' .. hex
		end
		hexadecimal = hexadecimal .. hex
	end
	return hexadecimal
end

function frame:UpdateOptions()
	-- General
	self:ClearAllPoints()
	self:SetPoint("CENTER", db.general.posx, db.general.posy)
	self:SetSize(db.general.width, db.general.height)
	-- Background --
	bgFrame.tex:SetVertexColor(db.background.color[1], db.background.color[2], db.background.color[3], db.background.color[4])
	-- Border --
	for i=1,4,1 do
		bordFrame.texs[i]:SetVertexColor(db.border.color[1], db.border.color[2], db.border.color[3], db.border.color[4])
		bordVfFrame.texs[i]:SetVertexColor(db.animations.bordercolor[1], db.animations.bordercolor[2], db.animations.bordercolor[3], db.animations.bordercolor[4])
	end
	SetBorderSize(bordFrame, db.border.thickness)
	SetBorderSize(bordVfFrame, db.border.thickness)
	-- Bars --
	local bartex = LSM:Fetch("statusbar", db.bars.texture)
	insBar:SetStatusBarTexture(bartex)
	insBar:SetStatusBarColor(db.bars.insbarcolor[1], db.bars.insbarcolor[2], db.bars.insbarcolor[3], 1)
	castBar:SetTexture(bartex)
	castBar:SetVertexColor(db.bars.castbarcolor[1], db.bars.castbarcolor[2], db.bars.castbarcolor[3], 1)
	asBar:SetTexture(bartex)
	asBar:SetVertexColor(db.bars.asbarcolor[1], db.bars.asbarcolor[2], db.bars.asbarcolor[3], 1)
	-- Enter Voidform Mark --
	lotvThreshold = db.misc.lotvthreshold
	local width = db.general.width
	local t = (db.overlay.thickness / 2)
	local offsetX = lotvThreshold / 100
	vfMarkFrame:ClearAllPoints()
	vfMarkFrame:SetPoint("TOPLEFT", insBar, "TOPLEFT", width * offsetX - t, 0)
	vfMarkFrame:SetPoint("BOTTOMRIGHT", insBar, "BOTTOMRIGHT", -width * (1 - offsetX) + t, 0)
	vfMarkFrame.tex:SetVertexColor(db.overlay.color[1], db.overlay.color[2], db.overlay.color[3], 1)
	-- Font --
	local file = LSM:Fetch("font", db.text.file)
	local flag = db.text.flag
	insFontString:SetFont(file, db.text.insanitysize, flag)
	insFontString:SetTextColor(db.text.insanitycolor[1], db.text.insanitycolor[2], db.text.insanitycolor[3], 1)
	vfTimeFontString:SetFont(file, db.text.vftimesize, flag)
	vfTimeFontString:SetTextColor(db.text.vftimecolor[1], db.text.vftimecolor[2], db.text.vftimecolor[3], 1)
	vfStackFontString:SetFont(file, db.text.vfstacksize, flag)
	vfStackFontString:SetTextColor(db.text.vfstackcolor[1], db.text.vfstackcolor[2], db.text.vfstackcolor[3], 1)
	insHexColor, castHexColor, asHexColor = RGBAToHex(db.text.castcolor), RGBAToHex(db.text.castcolor), RGBAToHex(db.text.ascolor)
	-- Animations --
	oocFadeAnim.alpha:SetDuration(db.animations.oocduration)
	oocFadeAnim.alpha:SetStartDelay(db.animations.oocstartdelay)
	oocFadeAnim.alpha:SetToAlpha(db.general.alphaooc)
	bordVfFrame:SetAlpha(0)
	vfReadyAnim.alphaIn:SetDuration(db.animations.borderduration / 2)
	vfReadyAnim.alphaOut:SetDuration(db.animations.borderduration / 2)
	-- Lingering Insanity --
	displayLiStacks = db.misc.lingeringinsanity
	-- STM reporting --
	vfDisplayRepEnabled = db.misc.vfdisreportingenabled
	vfSaveRepEnabled = db.misc.vfsavereportingenabled
	maxSavedReps	= db.misc.vfreportingmaxsaved
	-- GCD --
	gcdBarEnable = db.gcd.enable
	gcdBarDeplete = db.gcd.deplete
	gcdBar:ClearAllPoints()
	if db.gcd.pos == "TOP" then
		gcdBar:SetPoint("BOTTOMLEFT", bgFrame, "TOPLEFT", 0, db.border.thickness)
		gcdBar:SetPoint("BOTTOMRIGHT", bgFrame, "TOPRIGHT", 0, db.border.thickness)
	else
		gcdBar:SetPoint("TOPLEFT", bgFrame, "BOTTOMLEFT", 0, -db.border.thickness)
		gcdBar:SetPoint("TOPRIGHT", bgFrame, "BOTTOMRIGHT", 0, -db.border.thickness)
	end
	gcdBar:SetHeight(db.gcd.height)
	gcdBar:SetStatusBarTexture(bartex)
	gcdBar:SetStatusBarColor(db.gcd.barcolor[1], db.gcd.barcolor[2], db.gcd.barcolor[3], 1)
	gcdBar:SetMinMaxValues(0, 1)
	gcdBar:SetValue(0)
	gcdBar.tex:SetVertexColor(db.background.color[2], db.background.color[2], db.background.color[3], db.background.color[4])
	gcdBar:Hide()
	-- Patching --
	--[[enablePatch = db.misc.patch70105enable
	if enablePatch == true then
	end]]--
end

SLASH_VILINKASINSANITY1, SLASH_VILINKASINSANITY2 = "/vilinkasinsanity", "/vilins"

SlashCmdList["VILINKASINSANITY"] = function(message)
	if select(2, UnitClass("player")) == priestClass then
		if GetSpecialization() == SPEC_PRIEST_SHADOW then
			if message == "report" or message == "vf" then
				PrintSavedVfReps()
			elseif message == "opt" then
				local loaded, reason = LoadAddOn("VilinkasInsanityOptions")
				if not loaded then
					print(vilins .. ": Options module could not be loaded: " .. _G["ADDON_" .. reason])
					return
				end
				
				if UnitAffectingCombat("player") then
					print(vilins .. ": Options will open when you leave comat.")
					_G["VilinkasInsanityOptions"].showWhenOutOfCombat = true
				else
					_G["VilinkasInsanityOptions"]:Show()
				end
			else
				print(vilins)
				print("Use |cFF9370DB/vilins opt|r to acces options menu.")
				print("Use |cFF9370DB/vilins vf|r display archived voidform reports.")
			end
		else
			print(vilins .. ":|r Current specialization is not supported.")
		end
	end
end
